using System.Security.Cryptography.X509Certificates;
using Autofac;
using FitConnect.Encryption;
using FitConnect.Interfaces;
using FitConnect.Models;
using FitConnect.Services.Models;
using FitConnect.Services.Models.v1;
using Microsoft.Extensions.Logging;
using Newtonsoft.Json;
using Metadata = FitConnect.Models.Api.Metadata.Metadata;

namespace FitConnect;

/// <summary>
///     Fluent API for the FitConnect.Subscriber
/// </summary>
public class Subscriber : FitConnectClient,
    IFluentSubscriber,
    IFluentSubscriberWithSubmission {
    public Submission? Submission { get; private set; }

    public Subscriber(FitConnectEnvironment environment,
        string clientId, string clientSecret,
        string privateKeyDecryption,
        string privateKeySigning,
        string publicKeyEncryption,
        string publicKeySignatureVerification,
        IContainer container) : base(environment, clientId, clientSecret,
        container,
        privateKeyDecryption, privateKeySigning, publicKeyEncryption,
        publicKeySignatureVerification) {
        Encryption = new FitEncryption(privateKeyDecryption, privateKeySigning, publicKeyEncryption,
            publicKeySignatureVerification,
            container.Resolve<ILogger>());
    }

    public Subscriber(FitConnectEnvironment environment,
        string clientId, string clientSecret,
        string privateKeyDecryption,
        string privateKeySigning,
        string publicKeyEncryption,
        string publicKeySignatureVerification,
        ILogger? logger = null) : base(environment, clientId, clientSecret, logger,
        privateKeyDecryption, privateKeySigning) {
        Encryption = new FitEncryption(privateKeyDecryption, privateKeySigning, publicKeyEncryption,
            publicKeySignatureVerification, logger);
    }

    public void AcceptSubmission(string caseId, string submissionId, string destinationId) {
        CompleteSubmission(caseId, submissionId, destinationId, FinishSubmissionStatus.Accepted);
    }

    public void RejectSubmission(string caseId, string submissionId, string destinationId) {
        CompleteSubmission(caseId, submissionId, destinationId, FinishSubmissionStatus.Rejected);
    }

    public void ForwardSubmission(string caseId, string submissionId, string destinationId) {
        CompleteSubmission(caseId, submissionId, destinationId,
            FinishSubmissionStatus.Forwarded);
    }


    public void CompleteSubmission(string caseId, string submissionId, string destinationId,
        FinishSubmissionStatus status) {
        var eventName = GetEvent(status);

        var token =
            Encryption.CreateSecurityEventToken(submissionId, caseId, destinationId, eventName);

        var result = CasesService.FinishSubmission(caseId, token);
        Logger?.LogInformation(result ?? "");
    }


    /// <summary>
    ///     Receives the available submissions from the server
    /// </summary>
    /// <param name="destinationId"></param>
    /// <param name="skip"></param>
    /// <param name="take"></param>
    /// <returns></returns>
    public IEnumerable<SubmissionForPickupDto> GetAvailableSubmissions(string? destinationId = null,
        int skip = 0,
        int take = 100) {
        var submissionsResult = SubmissionService.ListSubmissions(destinationId, 0, 100).Result;

        // Creating a dictionary of destinationId to submissionIds from the REST API result
        return submissionsResult.Submissions;
    }


    /// <summary>
    ///     Receives a specific submission from the server.
    /// </summary>
    /// <param name="submissionId"></param>
    /// <returns></returns>
    public IFluentSubscriberWithSubmission RequestSubmission(string submissionId) {
        var submission = (Submission)SubmissionService.GetSubmission(submissionId);
        var (metaDataString, _, metaHash) = Encryption.Decrypt(submission.EncryptedMetadata);

        submission.Metadata =
            JsonConvert.DeserializeObject<Metadata>(metaDataString);

        if (submission.EncryptedData != null) {
            var (dataString, _, dataHash) = Encryption.Decrypt(submission.EncryptedData);
            submission.Data = dataString;
        }

        Submission = submission;
        return this;
    }


    /// <summary>
    ///     Reading attachments for a submission.
    /// </summary>
    /// <returns></returns>
    /// <exception cref="NotImplementedException"></exception>
    public IEnumerable<Attachment> GetAttachments() {
        // TODO add guard calls

        var attachments = new List<Attachment>();
        foreach (var id in Submission!.AttachmentIds) {
            var encryptedAttachment = SubmissionService.GetAttachment(Submission.Id, id);
            var (_, content, hash) = Encryption.Decrypt(encryptedAttachment);
            var attachmentMeta =
                Submission.Metadata.ContentStructure.Attachments.First(a => a.AttachmentId == id);

            attachments.Add(new Attachment(attachmentMeta, content));
        }

        Submission.Attachments = attachments;
        return attachments;
    }


    private string GetEvent(FinishSubmissionStatus state) {
        return state switch {
            FinishSubmissionStatus.Accepted =>
                "https://schema.fitko.de/fit-connect/events/accept-submission",
            FinishSubmissionStatus.Rejected =>
                "https://schema.fitko.de/fit-connect/events/reject-submission",
            FinishSubmissionStatus.Forwarded =>
                "https://schema.fitko.de/fit-connect/events/forward-submission",
            _ => throw new ArgumentException("Invalid state")
        };
    }

    public void AcceptSubmission(SubmissionForPickupDto submission) =>
        CompleteSubmission(submission, FinishSubmissionStatus.Accepted);

    public void RejectSubmission(SubmissionForPickupDto submission) =>
        CompleteSubmission(submission, FinishSubmissionStatus.Rejected);

    public void ForwardSubmission(SubmissionForPickupDto submission) =>
        CompleteSubmission(submission, FinishSubmissionStatus.Forwarded);

    public void CompleteSubmission(SubmissionForPickupDto submission,
        FinishSubmissionStatus status) {
        if (submission.SubmissionId == null || submission.CaseId == null ||
            submission.DestinationId == null) {
            throw new ArgumentException("Submission does not contain all required fields");
        }

        var eventName = GetEvent(status);

        var token =
            Encryption.CreateSecurityEventToken(submission.SubmissionId, submission.CaseId,
                submission.DestinationId, eventName);

        var result = CasesService.FinishSubmission(submission.CaseId, token);
        Logger?.LogInformation("Submission completed {status}", result);
    }
}

public enum FinishSubmissionStatus {
    Accepted,
    Rejected,
    Forwarded
}
